part of query;

class Foreign {
  /// The table in which foreign key references the column [col]
  final String table;

  /// References column in table [table]
  final String col;

  const Foreign(this.table, this.col);

  static Foreign make(String table, String refCol) =>
      new Foreign(table, refCol);
}

abstract class CreateColumn<ValType> implements ToSqlable {
  String get colName;

  bool get isNullable;

  bool get isPrimaryKey;

  Foreign get foreignKey;
}

class CreateIntColumn extends CreateColumn<int> {
  bool _nullable = false;

  String _colName;

  bool _autoIncrement = false;

  bool _primary = false;

  Foreign _foreignKey;

  CreateIntColumn(this._colName,
      {bool nullable: false,
      bool autoIncrement: false,
      bool primary: false,
      Foreign foreignKey})
      : _nullable = nullable,
        _autoIncrement = autoIncrement,
        _primary = primary,
        _foreignKey = foreignKey;

  CreateIntColumn.Primary(this._colName, {Foreign foreignKey})
      : _primary = true,
        _foreignKey = foreignKey;

  CreateIntColumn.AutoPrimary(this._colName)
      : _primary = true,
        _autoIncrement = true;

  CreateIntColumn nullable() {
    _nullable = true;
    return this;
  }

  CreateIntColumn foreign(String table, {String refCol}) {
    if (_foreignKey != null) throw new Exception("Foreign key alread set!");
    if (refCol == null) refCol = _colName;
    _foreignKey = new Foreign(table, refCol);
    return this;
  }

  String get colName => _colName;

  bool get isNullable => _nullable;

  bool get isPrimaryKey => _primary;

  Foreign get foreignKey => _foreignKey;

  String toSql() {
    String ret = _colName;
    if (_autoIncrement) {
      ret += ' SERIAL';
    } else {
      ret += ' INT';
    }
    if (!_nullable) ret += ' NOT NULL';
    return ret;
  }

  /* TODO
  String toSql() =>
      _colName +
      ' INT' +
      (_nullable ? ' NOT NULL' : '') +
      (_autoIncrement ? ' AUTO_INCREMENT' : '');
      */
}

class CreateDoubleColumn extends CreateColumn<int> {
  bool _nullable = false;

  String _colName;

  bool _autoIncrement = false;

  bool _primary = false;

  Foreign _foreignKey;

  CreateDoubleColumn(this._colName,
      {bool nullable: false,
      bool autoIncrement: false,
      bool primary: false,
      Foreign foreignKey})
      : _nullable = nullable,
        _autoIncrement = autoIncrement,
        _primary = primary,
        _foreignKey = foreignKey;

  CreateDoubleColumn.Primary(this._colName, {Foreign foreignKey})
      : _primary = true,
        _foreignKey = foreignKey;

  CreateDoubleColumn.AutoPrimary(this._colName)
      : _primary = true,
        _autoIncrement = true;

  CreateDoubleColumn nullable() {
    _nullable = true;
    return this;
  }

  CreateDoubleColumn foreign(String table, {String refCol}) {
    if (_foreignKey != null) throw new Exception("Foreign key alread set!");
    if (refCol == null) refCol = _colName;
    _foreignKey = new Foreign(table, refCol);
    return this;
  }

  String get colName => _colName;

  bool get isNullable => _nullable;

  bool get isPrimaryKey => _primary;

  Foreign get foreignKey => _foreignKey;

  String toSql() {
    String ret = _colName;
    ret += ' double precision';
    if (!_nullable) ret += ' NOT NULL';
    return ret;
  }

/* TODO
  String toSql() =>
      _colName +
      ' INT' +
      (_nullable ? ' NOT NULL' : '') +
      (_autoIncrement ? ' AUTO_INCREMENT' : '');
      */
}

class CreateBoolColumn extends CreateColumn<bool> {
  bool _nullable = false;

  String _colName;

  bool _primary = false;

  CreateBoolColumn(this._colName, {bool nullable: false, bool primary: false})
      : _nullable = nullable,
        _primary = primary;

  CreateBoolColumn nullable() {
    _nullable = true;
    return this;
  }

  String get colName => _colName;

  bool get isNullable => _nullable;

  bool get isPrimaryKey => _primary;

  Foreign get foreignKey => null;

  String toSql() => _colName + ' BOOLEAN' + (!_nullable ? ' NOT NULL' : '');
}

class CreateStrColumn extends CreateColumn<String> {
  bool _nullable = false;

  String _colName;

  bool _primary = false;

  int _length = 20;

  Foreign _foreignKey;

  CreateStrColumn(this._colName,
      {bool nullable: false,
      bool primary: false,
      int length: 20,
      Foreign foreignKey})
      : _nullable = nullable,
        _length = length,
        _primary = primary,
        _foreignKey = foreignKey;

  CreateStrColumn.Primary(this._colName, {int length: 20, Foreign foreignKey})
      : _primary = true,
        _length = length,
        _foreignKey = foreignKey;

  CreateStrColumn nullable() {
    _nullable = true;
    return this;
  }

  CreateStrColumn foreign(String table, {String refCol}) {
    if (_foreignKey != null) throw new Exception("Foreign key alread set!");
    if (refCol == null) refCol = _colName;
    _foreignKey = new Foreign(table, refCol);
    return this;
  }

  String get colName => _colName;

  bool get isNullable => _nullable;

  bool get isPrimaryKey => _primary;

  int get length => length;

  Foreign get foreignKey => _foreignKey;

  String toSql() =>
      _colName + ' VARCHAR(${_length})' + (!_nullable ? ' NOT NULL' : '');
}

class CreateStatement implements Statement {
  String _tableName;

  final Map<String, CreateColumn> _columns = {};

  bool _ifNotExists = false;

  CreateStatement();

  CreateStatement named(String tableName) {
    _tableName = tableName;
    return this;
  }

  CreateStatement ifNotExists([bool create = true]) {
    _ifNotExists = create;
    return this;
  }

  CreateStatement intCol(String colName,
      {bool autoIncrement: false,
      bool primary: false,
      String foreignTable,
      String foreignCol}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateIntColumn(colName,
        autoIncrement: autoIncrement, primary: primary, foreignKey: foreign);
    return this;
  }

  CreateStatement intNullCol(String colName,
      {bool primary: false,
      Foreign foreign,
      String foreignTable,
      String foreignCol}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateIntColumn(colName,
        nullable: true, primary: primary, foreignKey: foreign);
    return this;
  }

  CreateStatement doubleCol(String colName,
      {bool autoIncrement: false,
      bool primary: false,
      String foreignTable,
      String foreignCol}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateDoubleColumn(colName,
        autoIncrement: autoIncrement, primary: primary, foreignKey: foreign);
    return this;
  }

  CreateStatement doubleNullCol(String colName,
      {bool primary: false,
      Foreign foreign,
      String foreignTable,
      String foreignCol}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateDoubleColumn(colName,
        nullable: true, primary: primary, foreignKey: foreign);
    return this;
  }

  CreateStatement boolCol(String colName) {
    _columns[colName] = new CreateBoolColumn(colName);
    return this;
  }

  CreateStatement boolNullCol(String colName) {
    _columns[colName] = new CreateBoolColumn(colName).nullable();
    return this;
  }

  CreateStatement dateCol(String colName) {
    _columns[colName] = new CreateBoolColumn(colName);
    return this;
  }

  CreateStatement dateNullCol(String colName) {
    _columns[colName] = new CreateBoolColumn(colName).nullable();
    return this;
  }

  CreateStatement strCol(String colName,
      {int length: 20,
      bool primary: false,
      String foreignTable,
      String foreignCol}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateStrColumn(colName,
        length: length, primary: primary, foreignKey: foreign);
    return this;
  }

  CreateStatement strNullCol(String colName,
      {int length: 20,
      bool primary: false,
      String foreignTable,
      String foreignCol}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateStrColumn(colName,
            length: length, primary: primary, foreignKey: foreign)
        .nullable();
    return this;
  }

  CreateStatement column(CreateColumn column) {
    _columns[column.colName] = column;
    return this;
  }

  CreateStatement columns(List<CreateColumn> columns) {
    for (CreateColumn col in columns) {
      _columns[col.colName] = col;
    }
    return this;
  }

  String toSql() {
    StringBuffer sb = new StringBuffer();

    sb.write('CREATE TABLE');

    if (_ifNotExists) sb.write(' IF NOT EXISTS');

    sb.write(' $_tableName (');

    sb.write(_columns.values.map((CreateColumn col) => col.toSql()).join(', '));

    final List<CreateColumn> primaries =
        _columns.values.where((CreateColumn col) => col.isPrimaryKey).toList();
    if (primaries.length != 0) {
      sb.write(', PRIMARY KEY (');
      sb.write(primaries.map((CreateColumn col) => col.colName).join(','));
      sb.write(')');
    }

    for (CreateColumn col in _columns.values) {
      if (col.foreignKey != null) {
        sb.write(
            ', FOREIGN KEY (${col.colName}) REFERENCES ${col.foreignKey.table}(${col.foreignKey.col})');
      }
    }

    sb.write(')');

    return sb.toString();
  }
}
